# AlbumReview::Credits.pm by mh 2009
package Plugins::AlbumReview::Credits;

use strict;
use Plugins::AlbumReview::Plugin;
use Slim::Utils::Log;
use Slim::Utils::Prefs;
use Slim::Utils::Strings qw(string cstring);
use Slim::Networking::SimpleAsyncHTTP;

my $log = logger('plugin.albumreview');
my $prefs = preferences('plugin.albumreview');

my $baseURL    = 'http://www.allmusic.com';

sub loadCredits {
	my ($client, $params, $callback, $httpClient, $response, $request) = @_;
	my $status = \%Plugins::AlbumReview::Plugin::status;

	return unless $status->{$client}{creditsUrl};

	my $http = Slim::Networking::SimpleAsyncHTTP->new(\&gotAlbumCredits, \&Plugins::AlbumReview::Plugin::gotErrorViaHTTP, 
		{
			client => $client, 
			params => $params, 
			callback => $callback, 
			httpClient => $httpClient, 
			response => $response, 
			searchAlbum => {
				album => $status->{$client}{creditsUrl},
				allmusicUrl => $status->{$client}{creditsUrl},
				artist => 'allmusic',
			},
			cliRequest => $request,
			timeout => 30
		});

	$http->get($status->{$client}{creditsUrl});
}

sub gotAlbumCredits {
	my $http = shift;
	my $params = $http->params();

	my $cachedFile = Plugins::AlbumReview::Plugin::getCachedAlbumFilename($params->{searchAlbum});

	my $html = $http->content();
	if (not $html) {
		$log->warn("AlbumReview: timeout problem?");
	}
	else {
		Plugins::AlbumReview::Plugin::storeCachedFile($cachedFile, \$html);

		# CLI answer
		if (defined $params->{cliRequest} && $params->{cliRequest}) {
			if (getCachedCredits($params->{cliRequest}->client(), $params->{'params'})) {
				_cliPrepareResponse($params->{cliRequest});
			}
		}
	} 

}

sub getCachedCredits {
	my $client = shift;

	my $status = \%Plugins::AlbumReview::Plugin::status;
	$log->debug($status->{$client}{creditsUrl});
	
	if ($status->{$client}{creditsUrl} && defined $status->{$client}{credits}) {
		$log->info("AlbumReview (cache): credits $status->{$client}{creditsUrl} were in memory cache");	
	}
	else {
		# Try and use the locally cached copy first
		my $cachedFile = Plugins::AlbumReview::Plugin::getCachedAlbumFilename({
			album => $status->{$client}{creditsUrl},
			allmusicUrl => $status->{$client}{creditsUrl},
			artist => 'allmusic',
		});
		my $cacheAge = $prefs->get('plugin_albumreview_cache') || 30;

		# we're fine! cached file found
		if (-e $cachedFile && -M $cachedFile < $cacheAge) {
			$log->warn("credits were in disk cache ($status->{$client}{creditsUrl})");
			local $/ = undef;
	
			open CACHE, $cachedFile;
			my $creditsHTML = <CACHE>;
			close CACHE;
			
			my @credits;

			my ($creditsList) = $creditsHTML =~ /section class="credits">(.*?)<\/section/si;

			while ($creditsList =~ /<tr>.*?href="(.*?)".*?>\s*(.*?)\s*<.*?class="credit".*?>\s*(.*?)\s*<\/td/sig) {
				my $url = "$1/biography";
				my $contributor = $2;
				my $contribution = $3;
				
				push @credits, [$contributor, $url, $contribution];
			}

			$status->{$client}{credits} = \@credits;
		}
		else {
			
			$log->warn("AlbumReview (cache): need to look online for $status->{$client}{creditsUrl}...");
			$status->{$client}{credits} = undef;
			return 0;
		}
	}
	
	return 1;
}

sub cliQuery {
	my $request = shift;
	
	$log->warn("AlbumReview: cliCreditsQuery()");
	
	# check this is the correct query.
	if ($request->isNotQuery([['albumcredits'], ['get']]) || !$request->client()) {
		$request->setStatusBadDispatch();
		return;
	}
	
	my $client = $request->client();
	$request->setStatusProcessing();

	my $params = $request->getParamsCopy();

	my $review = Plugins::AlbumReview::Plugin::getCachedReview($client, $params);
	if ((not $review) && (not defined $Plugins::AlbumReview::Plugin::status{$client}{status})) {
		no strict 'subs';
		Plugins::AlbumReview::Plugin::loadReview($client, $params, undef, undef, undef, $request);
	}
	elsif ($review) {
		Plugins::AlbumReview::Plugin::cliPrepareResponse($request);
	}
}


sub cliPrepareResponse {
	my ($client, $request) = @_;
	
	my $status = \%Plugins::AlbumReview::Plugin::status;

	$log->info($status->{$client}{creditsUrl});
	
	if ($status->{$client}{creditsUrl}) {

		if (!getCachedCredits($client)) {
			loadCredits($client, undef, undef, undef, undef, $request);
		}
		else {
			_cliPrepareResponse($request);
		}
	}
	else {
		if ($request->getParam('menu')) {
			$request->addResult('window', {
				textArea => $client->string('PLUGIN_ALBUMREVIEW_NOREVIEW'),
			});	
		}
	}
}

sub _cliPrepareResponse {
	my ($request) = @_;
	
	my $status = \%Plugins::AlbumReview::Plugin::status;
	my $client = $request->client();

	$log->info("AlbumReview: _cliPrepareResponse");
	
	my $chunkCount = 0;
	
	if ($request->getParam('menu')) {
		my $canBio = Slim::Utils::PluginManager->isEnabled('Plugins::Biography::Plugin');

		foreach (@{$status->{$client}{credits}}) {
			# don't return more than 200 items
			last if $chunkCount == 200;

			if ($canBio) {
				$request->addResultLoop('item_loop', $chunkCount, 'actions', {
					go => {
						cmd => [ 'biography', 'get' ],
						params => {
							menu => 1,
							allmusicUrl => @$_[1],
							artist => @$_[0],
						}					
					}
				});
			}
			else {
				$request->addResultLoop('item_loop', $chunkCount, 'type', 'text');
				$request->addResultLoop('item_loop', $chunkCount, 'style', 'itemNoAction');
			}
			
			$request->addResultLoop('item_loop', $chunkCount, 'text', @$_[0] . (@$_[2] ? " (@$_[2])" : ''));
			$chunkCount++;
		}
		$request->addResult('offset', 0);
		$request->addResult('count', $chunkCount);
	}
	else {
		foreach (@{$status->{$client}{credits}}) {
			$request->addResultLoop('item_loop', $chunkCount, 'contributor', @$_[0] . (@$_[2] ? " (@$_[2])" : ''));
			$chunkCount++;
		}
	}
	
	$request->setStatusDone();
}


sub trackInfoHandler {
	my $return = _objInfoHandler( @_, 'track' );
	return $return;
}

sub albumInfoHandler {
	my $return = _objInfoHandler( @_, 'album' );
	return $return;
}

sub _objInfoHandler {
	my ( $client, $url, $obj, $remoteMeta, $tags, $objType ) = @_;

	my $album_id = $obj->id;
	if ( $objType eq 'track') {
		$album_id = $obj->album ? $obj->album->id : undef;
	}

	return {
		type   => 'redirect',
		name   =>  cstring($client, 'PLUGIN_ALBUMCREDITS'),

#		player => {
#			mode => 'Plugins::AlbumReview::Plugin',
#			modeParams => {
#				album_id => $album_id,
#			},
#		},
#
#		web  => {
#			url => 'plugins/AlbumReview/index.html?album_id=' . $album_id,
#		},
		
		jive => {
			actions => {
				go => {
					cmd => [ 'albumcredits', 'get' ],
					params => {
						menu => 1,
						album_id => $album_id,
					}					
				}
			}
		}
	};
}


1;