# AlbumReview::Plugin.pm by mh 2006
#
# This code is derived from code with the following copyright message:
#
# SliMP3 Server Copyright (C) 2001 Sean Adams, Slim Devices Inc.
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License,
# version 2.
#
# Changelog
# 2.9.9 - parser again
# 2.9.8 - fix Credits menu
#       - allmusic layout changes...
# 2.9.7 - fix compatibility with UEML 10.0
#       - allmusic layout changes...
# 2.9.6 - allmusic.com layout changes (credits still not working)
# 2.9.5 - fix cache initialization
#       - more parser fixes
# 2.9.4 - load /review directly if possible
# 2.9.2/3 - fix Allmusic page parsing
# 2.9.1 - add support for optional artist name when searching for an album
# 2.9.0 - add AlbumReview to Global Search
# 2.8.3 - fix "not found" issue, where this message wouldn't show on SP
#       - fix album name cleanup issue, where appendixes in brackets etc. wouldn't be ignored
#       - add 7.5/noweb compatibility
# 2.8.2 - fix SBS 7.4 compatibility
# 2.8.1 - improve compatibility with 7.4/TinySC
# 2.8.0 - add Album Credits feature, linking to Biography
#       - add support for SC 7.4's AlbumInfo menu
# 2.7.0 - support browsing alternate album list if no unique is found using the menu CLI (iPeng, SqueezePlay)
# 2.6.0 - SlimServer 7.1 compatibility
#       - add support for trackinfo menu
# 2.5.0 - SlimServer 7.0 compatibility
# 2.1.1 - improve recovery from network errors
# 2.1.0 - get rid of display stuff inside display stuff
# 2.0.2 - improve error handling on timeouts & cie.
#       - set initial state "loading..." (0), not "nothing found" (-1)
# 2.01 - further integrate screensaver and normal display mode
# 2.0  - if HTML::Formatter is not available, just don't display menu option
#      - add screensaver mode
# 1.99 - improved slimserver 6.5 compliance
#      - replace different html files for skins with one universal html file
#      - add CLI support ("00:00:00:00:00:00 albumreview get")
#      - fix an issue with looping when returning from the screensaver
#      - fix an issue with looping when AsyncHTTP or allmusic.com returned error
# 1.32 - fix a crasher when a http client is used
# 1.31 - make album/artist search case insensitive
# 1.30 - add AMG rating to the web page
# 1.22 - make better use of the Template Toolkit
# 1.21 - added option to remove stuff in brackets from album name
# 1.2  - recover from ip blocking and error messages (timeout) by allmusic 
# 1.1  - added publishing year
#      - improve error messages
#      - improved filter for "cd x of y" & co.
# 1.0  - first release

package Plugins::AlbumReview::Plugin;
use base qw(Slim::Plugin::Base);

use strict;
use Exporter;
our @EXPORTS = qw(setStatus getStatus);

use vars qw($VERSION);

use Digest::MD5 qw(md5_hex);
use File::Spec;
use FindBin qw($Bin);
use Slim::Utils::Strings qw(string cstring);
use Slim::Utils::Log;
use Slim::Utils::Prefs;
use Slim::Utils::Versions;
use Plugins::AlbumReview::Credits;

my $prefs = preferences('plugin.albumreview');

my $baseURL    = 'http://www.allmusic.com';
my $searchURL  = "$baseURL/search/albums";
my $albumRatingURL = "$baseURL/i/pages/site/icons/mini-amg.gif";
my $albumPickURL = "$baseURL/i/pages/site/icons/check.gif";

my $log = Slim::Utils::Log->addLogCategory({
	'category'     => 'plugin.albumreview',
	'defaultLevel' => 'ERROR',
	'description'  => 'PLUGIN_ALBUMREVIEW',
});

# TODO: move to $client->modeParams
our %status;

my $hasModules;

sub getDisplayName {
	return 'PLUGIN_ALBUMREVIEW';
}

sub initPlugin {
	my $class = shift;
	$class->SUPER::initPlugin(shift);

	$VERSION = $class->_pluginDataFor('version');

	if (Slim::Utils::Versions::compareVersions($::VERSION, '7.5') == -1 || !$::noweb) {
		require Plugins::AlbumReview::Settings;
		Plugins::AlbumReview::Settings->new;
	}

	eval {
		use lib(File::Spec->catdir($Bin, 'Plugins', 'MHCPAN'));
		require HTML::FormatText;
		$hasModules = 1;
	};
	
	if (Slim::Utils::Versions::compareVersions($::VERSION, '7.1') >= 0) {
		Slim::Menu::TrackInfo->registerInfoProvider( albumreview => (
			func => \&trackInfoHandler,
		) );

		Slim::Menu::TrackInfo->registerInfoProvider( albumcredits => (
			func => \&Plugins::AlbumReview::Credits::trackInfoHandler,
		) );
	}

	if (Slim::Utils::Versions::compareVersions($::VERSION, '7.4') >= 0) {
		Slim::Menu::AlbumInfo->registerInfoProvider( albumreview => (
			func => \&albumInfoHandler,
		) );

		Slim::Menu::AlbumInfo->registerInfoProvider( albumcredits => (
			func => \&Plugins::AlbumReview::Credits::albumInfoHandler,
		) );
	}

	if (Slim::Utils::Versions::compareVersions($::VERSION, '7.5') >= 0) {
		Slim::Menu::GlobalSearch->registerInfoProvider( albumreview => (
			func => \&searchHandler,
		) );
	}

	initSetting('plugin_albumreview_remove_brackets', 0);
	initSetting('plugin_albumreview_refresh', 0);
	initSetting('plugin_albumreview_cache', 999999);	

#                                                                    |requires Client
#                                                                    |  |is a Query
#                                                                    |  |  |has Tags
#                                                                    |  |  |  |Function to call
#                                                                    C  Q  T  F
	Slim::Control::Request::addDispatch(['albumreview', 'get'], [1, 1, 1, \&cliQuery]);
	Slim::Control::Request::addDispatch(['albumcredits', 'get'], [1, 1, 1, \&Plugins::AlbumReview::Credits::cliQuery]);
}

sub initSetting {
	my $setting = shift;
	my $default = shift;

	if (!defined $prefs->get($setting)) {
		$prefs->set($setting, $default);
	}
}

sub setMode {
	my ($class, $client, $method) = @_;

	if ($method eq 'pop') {
		Slim::Buttons::Common::popMode($client);
		return;
	}

	$status{$client}{status} = '0';
	$client->lines( \&lines );

	if ($client->modeParam('album_id')) {
		initReview($client, { album_id => $client->modeParam('album_id') });
	}
	else {
		initReview($client);
	}
}

sub screenSaver {
	Slim::Buttons::Common::addSaver( 'SCREENSAVER.albumreview',
		getScreensaverFunctions(), \&setScreensaverMode, 
		\&leaveScreensaverMode, 'PLUGIN_ALBUMREVIEW');
}

sub getScreensaverFunctions {
	my %functions = (
		'done' => sub {
			my ( $client, $funct, $functarg ) = @_;
			Slim::Buttons::Common::popMode($client);
			$client->update();
	
			# pass along ir code to new mode if requested
			if ( defined $functarg && $functarg eq 'passback' ) {
				Slim::Hardware::IR::resendButton($client);
			}
		}
	);

	return \%functions;
}

sub setScreensaverMode {
	my $client = shift;

	Slim::Control::Request::subscribe(\&updateReview, [['playlist'], ['newsong']]);

	$client->modeParam('modeUpdateInterval', 1);
	$client->lines( \&screensaverLines );

	$status{$client}{status} = '0';
	initReview($client);
}

sub leaveScreensaverMode {
	Slim::Control::Request::unsubscribe(\&updateReview);
}

sub screensaverLines {
	my $client = shift;

	my $lines = lines($client);

	if ($status{$client}{status} ne '0'
		&& $status{$client}{status} ne '-1'
		&& $status{$client}{status} ne '99'
		&& $status{$client}{status} ne '2') {

		return {
			'line' => [ $lines->{'line'}->[0], join(' ', @{$status{$client}{playerReview}}) ],
		};	
	}
	else {
		return $lines;
	}
}

sub initReview {
	my $client = shift;
	my $params = shift;
	
	return if (!defined $client);

	$status{$client}{line} ||= 0;

	my $review;
	unless ($status{$client}{status} eq '-1' || $status{$client}{status} eq '99') {
		$review = getCachedReview($client, $params);
	}

	if ((not $review) && (not defined $status{$client}{status})) {
		$client->block({
			'line' => [	
				$client->string('PLUGIN_ALBUMREVIEW'), 
				$client->string('PLUGIN_ALBUMREVIEW_COLLECTINGINFO')
			]
		});
		loadReview($client, $params);
	}
	else {
		$client->unblock;
	}
}

sub updateReview {
	my $request = shift;
	$log->debug("AlbumReview: updateReview");
	initReview($request->client());
}

sub lines {
	my $client = shift;

	my $line1 = '';
	my $line2 = '';

	if ($status{$client}{status} eq "0") {
		$line1 = $client->string('PLUGIN_ALBUMREVIEW');
		$line2 = $client->string('PLUGIN_ALBUMREVIEW_COLLECTINGINFO');
	} 
	elsif ($status{$client}{status} eq "-1" || $status{$client}{status} eq "99") {
		$line1 = $client->string('PLUGIN_ALBUMREVIEW');
		$line2 = $client->string('PLUGIN_ALBUMREVIEW_NOTHINGFOUND');
	} 
	elsif ($status{$client}{status} eq "2") {
		$line1 = $client->string('PLUGIN_ALBUMREVIEW');
		$line2 = $client->string('PLUGIN_ALBUMREVIEW_NOALBUM');
	} 
	else {
		$line1 = $client->string('PLUGIN_ALBUMREVIEW') . ' - ' . $status{$client}{album};
		$line2 = $status{$client}{playerReview}[$status{$client}{line}];
	}

	return { 'line' => [ $line1, $line2 ] };
}

sub getFunctions {
	my %functions = (
		'up' => sub  {
			my $client = shift;
			$status{$client}{line} = Slim::Buttons::Common::scroll($client, -1, $#{ $status{$client}{playerReview} } + 1, $status{$client}{line} || 0);
			$client->pushUp();
		},
	
		'down' => sub  {
			my $client = shift;
			$status{$client}{line} = Slim::Buttons::Common::scroll($client, 1, $#{ $status{$client}{playerReview} } + 1, $status{$client}{line} || 0);
			$client->pushDown();
		},
		
		'left' => sub  {
			my $client = shift;
			initStatus($client, '-1');
			Slim::Buttons::Common::popModeRight($client);
		},
	
		'right' => sub  {
			my $client = shift;
			$client->bumpRight();
		}
	);

	return \%functions;
}

sub webPages {
	my $class = shift;
	
	return if Slim::Utils::Versions::compareVersions($::VERSION, '7.5') >= 0 && $::noweb;
	
	my $done = 0;
	if (Slim::Utils::Versions::compareVersions($::VERSION, '7.4') >= 0) {
		eval {
			Slim::Web::Pages->addPageFunction("plugins/AlbumReview/index\.html", \&handleWebIndex);
			Slim::Web::Pages->addPageFunction("plugins/AlbumReview/index\.txt", \&handleWebIndex);
			Slim::Web::Pages->addPageFunction("plugins/AlbumReview/pure\.html", \&handleWebIndex);
			$done = 1;
		}
	}

	if (!$done) {
		Slim::Web::HTTP::addPageFunction("plugins/AlbumReview/index\.html", \&handleWebIndex);
		Slim::Web::HTTP::addPageFunction("plugins/AlbumReview/index\.txt", \&handleWebIndex);
		Slim::Web::HTTP::addPageFunction("plugins/AlbumReview/pure\.html", \&handleWebIndex);
	}
	
	Slim::Web::Pages->addPageLinks('plugins', { 'PLUGIN_ALBUMREVIEW' => 'plugins/AlbumReview/index.html' });
	Slim::Web::Pages->addPageLinks('pluginsiPeng', { 'PLUGIN_ALBUMREVIEW' => 'plugins/AlbumReview/index.html' });
}

sub handleWebIndex {
	my ($client, $params, $callback, $httpClient, $response) = @_;

	# some browsers (Ewe...) do too much encoding...
	if ($params->{'url_query'} =~ /player%3D/) {
		$params->{'url_query'} = Slim::Utils::Misc::unescape($params->{'url_query'});

		foreach my $param (split /\&/, $params->{'url_query'}) {
			if ($param =~ /([^=]+)=(.*)/) {
				my $name  = Slim::Utils::Misc::unescape($1, 1);
				my $value = Slim::Utils::Misc::unescape($2, 1);

				$params->{$name} = $value;
				$log->warn("HTTP parameter $name = $value");
			}
		}

		$client = Slim::Player::Client::getClient($params->{'player'}) if $params->{'player'};
		$params->{'myClientState'} = $client;
	}

	if (!$client) {
		$params->{'artist'} = '42';
		$params->{'album'} = string('NO_PLAYER_FOUND');
		$params->{'albumreview'} = string('NO_PLAYER_DETAILS');
	} 
	elsif ($status{$client}{status} eq '99') {
		$params->{'msg'} = string('PLUGIN_ALBUMREVIEW_NOTREADY') . ' ('. $status{$client}{status} . ')';
		$status{$client}{status} = 0;
	}
	elsif ((not getCachedReview($client, $params)) && (not defined $status{$client}{status})) {
		loadReview($client, $params, $callback, $httpClient, $response);
		return undef;
	}
	elsif (getCachedReview($client, $params)) {
		$params->{'artist'}       = $status{$client}{status};
		$params->{'album'}        = $status{$client}{album};
		$params->{'albumreview'}  = $status{$client}{htmlReview};
		$params->{'albumimages'}  = $status{$client}{images} if (not $status{$client}{notFound});
		$params->{'albumtracks'}  = $status{$client}{trackList};
		$params->{'albumyear'}    = $status{$client}{year};
		$params->{'albumrating'}  = $status{$client}{rating};

		if ($prefs->get('plugin_albumreview_refresh')) {
			$params->{'refresh'} = $prefs->get('plugin_albumreview_refresh') * 1000;
		}
	} 
	elsif ($status{$client}{status} eq '-1') {
		$params->{'msg'} = string('PLUGIN_ALBUMREVIEW_NOTREADY') . ' (' . $status{$client}{status} . ')';
		$params->{'refresh'} = 5000;
	}
	else {
		$params->{'artist'} = string('PLUGIN_ALBUMREVIEW_NOTHINGFOUND') . ' (' . $status{$client}{status} . ')';
	}


	# return text file with line breaks
	if ($params->{path} =~ /\.txt$/ && $params->{'albumreview'}) {
		$params->{'albumreview'} = stripCLIResponse($params->{'albumreview'}, 'html');	
	}

	return Slim::Web::HTTP::filltemplatefile($params->{'path'}, $params);
}

sub loadReview {
	my ($client, $params, $callback, $httpClient, $response, $request) = @_;
	$status{$client}{status} = '0';

	my $album;
	# shortcut if we already have the URL
	if (defined($params) && $params->{'album'} && $params->{'allmusicUrl'}) {
		$album = {
			album => $params->{'album'},
			allmusicUrl => $params->{'allmusicUrl'},
			artist => 'allmusic',
		};
	}
	elsif (defined($params) && exists $params->{'album_id'}) {
		$album = getAlbumFromAlbumId($client, $params->{'album_id'});
	}
	elsif (defined($params) && exists $params->{'album'}) {
		$album = {
			'artist' => '',
			'album' => cleanupAlbumName($params->{'album'}),
			'coverart' => undef
		}
	}
	else {
		$album = getAlbumFromSongURL($client);
	}

	my $http = Slim::Networking::SimpleAsyncHTTP->new(\&gotAlbumInfo, \&gotErrorViaHTTP, 
				{
					client => $client, 
					params => $params, 
					callback => $callback, 
					httpClient => $httpClient, 
					response => $response, 
					searchAlbum => $album,
					cliRequest => $request,
					timeout => 30
				});

	my $searchAlbum = Slim::Utils::Text::matchCase($album->{album} . ' ' . ($album->{artist} || ''));
	$searchAlbum = Slim::Utils::Misc::escape($searchAlbum);
	
	if ($params && $params->{'allmusicUrl'}) {
		$http->get($params->{'allmusicUrl'});
	}
	else {
		main::DEBUGLOG && $log->debug("$searchURL/$searchAlbum");
		$http->get("$searchURL/$searchAlbum");
	}
}

sub gotAlbumInfo {
	my $http = shift;
	my $params = $http->params();

   	initStatus($params->{client}, '-1');
	my $cachedFile = getCachedAlbumFilename($params->{searchAlbum});

	my $html = $http->content();

	#main::DEBUGLOG && warn Data::Dump::dump($html);
	
	if (not $html) {
		$log->warn("AlbumReview: timeout problem?");
		initStatus($params->{client}, '2', [string('PLUGIN_ALBUMREVIEW_NOTHINGFOUND')]);
	}
	
	# direct hit?
	elsif (($html =~ /Overview/si) && ($html =~ /Review/si) && ($html =~ /Credits/si)) {
		$log->warn("AlbumReview: Match!");
		storeCachedFile($cachedFile, \$html);
		initStatus($params->{client}, '0');
	} 
	
	# try artist & album combo
	else {
		my %albumsFound;
		my ($albumList) = $html =~ /<ul class="search-results.*?>(.*?)<\/ul/si;
		
		#main::DEBUGLOG && $log->debug($albumList);
		
		while ($albumList =~ m|<li class="album".*?>(.*?)</li|sig) {
			my $albumInfo = $1;

			main::DEBUGLOG && $log->debug($albumInfo);

			my ($artist) = $albumInfo =~ /div class="artist".*?href=".*?".*?>\s*(.*?)\s*</sig;
			my ($url, $album) = $albumInfo =~ /div class="title".*? href="(.*?)".*?>(.*?)</sig;
			my ($year) = $albumInfo =~ /div class="year.*?(\d{4})/sig;
			
			$artist = lc($artist);
			$album  = lc($album);
			
			main::DEBUGLOG && $log->debug("$album, $artist, $year, $url");

			$albumsFound{$artist}->{$album} = {
				'url' => $url, 
				'year' => $year,
			} unless $albumsFound{$artist}->{$album} || !$artist || !$album;
		}
		
		# let's see if we've got our artist in the current list
		my $albumTitle = escapeRegExpChars($params->{searchAlbum}->{album});
		
		my $searchAlbumArtist = lc($params->{searchAlbum}->{artist});
		my $searchAlbum = lc($params->{searchAlbum}->{album});

		main::DEBUGLOG && $log->debug(Data::Dump::dump(%albumsFound, $searchAlbumArtist, $searchAlbum));

		if ($albumsFound{$searchAlbumArtist}->{$searchAlbum}) {
			# got it! Get that page
			$log->warn("AlbumReview: direct hit! $params->{searchAlbum}->{album}: $params->{searchAlbum}->{artist}, $albumsFound{$searchAlbumArtist}->{$searchAlbum}->{year}");
			my $http = Slim::Networking::SimpleAsyncHTTP->new(\&gotMoreAlbumInfo, \&gotErrorViaHTTP, $params);
			my $url = $albumsFound{$searchAlbumArtist}{$searchAlbum}->{url};
			$http->get($url);
			# don't create the page yet!
			return;
		}
		elsif (my @albums = grep(/$albumTitle/i, keys %{$albumsFound{$searchAlbumArtist}})) {
			# we got the artist, see if we can find a similar album
			$log->warn("AlbumReview: artist found, but not album. $params->{searchAlbum}->{album}: $params->{searchAlbum}->{artist}");
			
			# are we lucky? give the first a try...
			my $http = Slim::Networking::SimpleAsyncHTTP->new(\&gotMoreAlbumInfo, \&gotErrorViaHTTP, $params);
			my $url = $albumsFound{$searchAlbumArtist}{$albums[0]}->{url};
			$url .= '/review' if $url !~ /\/review$/;

			$log->warn("AlbumReview: $url");
			$http->get($url);
			# don't create the page yet!
			return;			
		}
		else {
			$log->warn("AlbumReview: nothing found?");
			# got nothing at all... show list
			storeCachedFile($cachedFile, \"<!-- NOT FOUND --><ul>$html</ul>\n");
			initStatus($params->{client}, '0');
		}
	}
	createAsyncResponse($params);
	$log->warn("AlbumReview: gotAlbumInfo");
	initReview($params->{'client'}, $params->{'params'});
}

sub gotMoreAlbumInfo {
	my $http = shift;
	my $html = $http->content();
	my $params = $http->params();
	$log->warn("AlbumReview: gotMoreAlbumInfo");
	storeCachedFile(getCachedAlbumFilename($params->{searchAlbum}), \$html);
	initStatus($params->{client}, '0');
	createAsyncResponse($params);
	initReview($params->{'client'}, $params->{'params'});
}

sub gotErrorViaHTTP {
	my $http = shift;
	$log->warn(Data::Dump::dump($http));
	my $params = $http->params();
	initStatus($params->{client}, '99', [string('PLUGIN_ALBUMREVIEW_NOTREADY')]);
	createAsyncResponse($params);
	initReview($params->{'client'}, $params->{'params'});
}

sub createAsyncResponse {
	my $params = shift;
	# create webpage if we were called by the web interface

	no strict qw/refs subs/;
	if (defined $params->{httpClient} && $params->{httpClient}) {
		my $output = handleWebIndex($params->{client}, $params->{params}, $params->{callback}, $params->{httpClient}, $params->{response});
		my $current_player;
		if (defined($params->{client})) {
			$current_player = $params->{client}->id();
		}
		
		$params->{callback}->($current_player, $params->{params}, $output, $params->{httpClient}, $params->{response});
	}
	
	# CLI answer
	if (defined $params->{cliRequest} && $params->{cliRequest}) {
		if (getCachedReview($params->{cliRequest}->client(), $params->{'params'})) {
			cliPrepareResponse($params->{cliRequest});
		}
	}

	$log->warn("AlbumReview: page created?");
}


sub getCachedReview {
	my $client = shift;
	my $params = shift;

	# check whether there's actually a song!
	my $album;
	# shortcut if we already have the URL
	if (defined($params) && $params->{'album'} && $params->{'allmusicUrl'}) {
		$log->debug("got URL: $params->{'allmusicUrl'}");
		$album = {
			album => $params->{'album'},
			allmusicUrl => $params->{'allmusicUrl'},
			artist => 'allmusic',
		};
	}
	elsif (defined($params) && exists $params->{'album_id'}) {
		$log->debug("got album from ID: $params->{'album_id'}");
		$album = getAlbumFromAlbumId($client, $params->{'album_id'});
	}
	elsif (defined($params) && exists $params->{'album'}) {
		my $artist = $params->{'artist'} || '';
		$album = {
			'artist' => $artist,
			'album' => cleanupAlbumName($params->{'album'}),
			'coverart' => undef
		}
	}
	else {
		$album = getAlbumFromSongURL($client);
		$log->debug("got album/artist from URL: $album->{album}, $album->{artist}");
	}

	if ($album && defined $album->{album} && defined $album->{artist}) {
		# only read review if artist and album changed
		if ((defined $status{$client}{status}) && ($album->{artist} eq $status{$client}{status}) 
				&& (defined $status{$client}{album}) && ($album->{album} eq $status{$client}{album})) {
			$log->info("AlbumReview (cache): $album->{album}/$album->{artist} was in memory cache");	
		}
		else {
			# initialize status hash
			initStatus($client, '-1');

			# Try and use the locally cached copy first
			my $cachedFile = getCachedAlbumFilename($album);
			my $cacheAge = $prefs->get('plugin_albumreview_cache') || 30;

			# we're fine! cached file found
			if (-e $cachedFile && -M $cachedFile < $cacheAge) {
				$log->warn("AlbumReview (cache): $album->{album}/$album->{artist} was in disk cache");
				local $/ = undef;
		
				open CACHE, $cachedFile;
				my $albumHTML = <CACHE>;
				close CACHE;
				
				# cache values for web interface
				$status{$client}{status} = $album->{artist};
				$status{$client}{album} = $album->{album};

				if ((not $albumHTML) || $albumHTML =~ /(traffic monitoring.*?allmusic.*?rate and speed|Error 2 Timeout)/i) {
					$log->warn("AlbumReview: Oops... Timeout? Or have we been blocked by allmusic.com?..");
					
					$status{$client}{htmlReview} = string('PLUGIN_ALBUMREVIEW_NOTREADY');
					$status{$client}{playerReview} = [$client->string('PLUGIN_ALBUMREVIEW_NOREVIEW')];

					# remove the file as it only contents an error message
					unlink $cachedFile;
				}
				elsif ($albumHTML =~ /<!-- NOT FOUND -->/i) {
					$log->warn("AlbumReview: Display albumlist");
					
					$status{$client}{htmlReview} = '';

					while ($albumHTML =~ /<tr class="significant".*?>(.*?)<\/tr>/sig) {
						my $albumEntry = $1;
						my @albumInfo = split(/<td.*?>(.*?)<\/td.*?>/si, $albumEntry);
						
						# get the album name and the url to its review
						$albumInfo[5] =~ /<a href="(\w+?)">(.*?)<\/a>/si;

						$log->warn("AlbumReview: $albumInfo[5] ($albumInfo[7], $albumInfo[11])");
						$status{$client}{htmlReview} .= '<li>' . fixBaseURL($albumInfo[5]) . " - ($albumInfo[7], $albumInfo[11])</li>\n";
					}
					
					# remove other artists, if our artist is in list
					my $albumTitle = escapeRegExpChars($album->{album});
					my $artist = escapeRegExpChars($album->{artist});
					if (($status{$client}{htmlReview} =~ /$artist/i) || ($status{$client}{htmlReview} =~ /$albumTitle/i)) {
						my $htmlReview = '';
						while ($status{$client}{htmlReview} =~ /(<li>.*?<\/li>)/sig) {
							my $line = $1;
							$htmlReview .= $line if (($line =~ /$artist/i) || ($line =~ /$albumTitle/i));
						}
						$status{$client}{htmlReview} = $htmlReview;
					}
					
					$status{$client}{htmlReview} = "<ul>$status{$client}{htmlReview}</ul>";
					$status{$client}{playerReview} = [$client->string('PLUGIN_ALBUMREVIEW_NOREVIEW')];
				}
				else {
					$status{$client}{htmlReview} = getAlbumReview(\$albumHTML);
					$status{$client}{trackList} = getTrackList(\$albumHTML);
					$status{$client}{year} = getAlbumYear(\$albumHTML);
					$status{$client}{rating} = getAlbumRating(\$albumHTML);
					$status{$client}{creditsUrl} = getAlbumCreditsLink(\$albumHTML);
					
					if ($status{$client}{htmlReview} eq $client->string('PLUGIN_ALBUMREVIEW_NOREVIEW')) {
						$log->warn("AlbumReview: Review found, but empty...");
						$status{$client}{playerReview} = [$client->string('PLUGIN_ALBUMREVIEW_NOREVIEW')];
						$status{$client}{notFound} = 1;
					}
					# display a list of found albums
					else {
						$status{$client}{images} = $album->{coverart};
						my $formattedReview = $status{$client}{htmlReview};
	
						if ($hasModules) {
							# wrap for player's display
							my $reviewFormatter = HTML::FormatText->new(leftmargin => 0, rightmargin => (charsPerLine($client, $formattedReview)-1));
							$formattedReview = $reviewFormatter->format_string($formattedReview);
							$formattedReview =~ s/\n+/\n/g;
			
							@{$status{$client}{playerReview}} = split(/\n/, $formattedReview);
						}
						else {
							push(@{$status{$client}{playerReview}}, $client->string('PLUGIN_ALBUMREVIEW_MISSING_MODULES'));
						}

						if (@{$status{$client}{playerReview}} == 0) {
							push(@{$status{$client}{playerReview}}, $client->string('PLUGIN_ALBUMREVIEW_NOTEXTFOUND'));
						} 
						else {
							my $rating = getAlbumRating(\$albumHTML, $client);
							if ($rating) {
								unshift(@{$status{$client}{playerReview}}, $rating);
							}
							push(@{$status{$client}{playerReview}}, '------------------------------------------------------------------------');
						}
					}
				}
			}
			# no cached content - we'll have to go the hard way...
			else {
				$log->warn("AlbumReview (cache): need to look online for $album->{artist}/$album->{album}...");
				initStatus($client, undef);
				return 0;
			} 
		}
	} 
	else {
		$log->warn("AlbumReview (cache): no album found");
		initStatus($client, 2, [string('PLUGIN_ALBUMREVIEW_NOALBUM')]);
		$status{$client}{album} = string('PLUGIN_ALBUMREVIEW_NOALBUM');
		return 0;
	}
	return 1;
}

# calculate the average character width for the current display based on our text...
sub charsPerLine {
	my $client = shift;
	my $text = shift;
	
	my $width = 40;

	# wrap the text to the device's width
	if (UNIVERSAL::can($client, 'displayWidth') && $client->displayWidth() > 40) {
		$text =~ s/\n//gs;
		$width = int($client->displayWidth / ($client->measureText($text, 2) / length($text))) - 1;
	}
	return $width;
}

sub storeCachedFile {
	my $cachedFile    = shift;
	my $cachedContent = shift;

	$log->warn("AlbumReview: writing '$cachedFile' to cache");
	open(CACHE, ">", $cachedFile) || $log->warn("Could not open $cachedFile for writing: $!");
	print CACHE $$cachedContent;
	close(CACHE);
}

sub getCachedAlbumFilename {
	my $album = shift;

	my $artistName = Slim::Utils::Misc::escape($album->{artist} || '');
	my $albumName = Slim::Utils::Misc::escape($album->{album});
	my $cacheFolder = Plugins::AlbumReview::Settings->getCacheFolder();
	
	my $filename = "$cacheFolder/" . md5_hex("$artistName-$albumName") . ".html"; 
	$log->is_debug && $log->debug("$artistName-$albumName: $filename");
	return $filename;
}

sub getAlbumReview {
	my $albumHTML = shift;
	if ($$albumHTML =~ /class="text.*?>(.*?)<\/div/si) {
		my $review = $1;
		$$albumHTML =~ /.*class="review-author.*?">(.*?)</si;
		
		return fixBaseURL("$review<br><br>($1)");
	}

	return string('PLUGIN_ALBUMREVIEW_NOREVIEW');
}

sub getAlbumCreditsLink {
	my $albumHTML = shift;

	if ($$albumHTML =~ /credits">.*?href="(.*?)"/sig) {
		return  $baseURL . $1;
	}
}

sub getTrackList {
	my $albumHTML = shift;
	my %trackList;
	
	my ($trackList) = $$albumHTML =~ / class="track-listing">(.*?)<\/section/si;

	while ($trackList =~ /tr class="track.*?">.*?"tracknum">.*?(\d+).*?<.*?title".*?>(.*?)<div.*?composer">(.*?)<div.*?"time">(.*?)</sig) {
		my ($trackNo, $title, $composer, $time) = ($1, $2, $3, $4);
		
		my %trackInfo = ( 
			'title' => fixBaseURL($title),
			'composer' => fixBaseURL($composer),
			'time' => $time
		);

		# pad the tracknumbers to get them correctly sorted in TT
		$trackList{sprintf('%3.d', $trackNo)} = \%trackInfo;
	}

	return \%trackList;
}

sub getAlbumYear {
	my $albumHTML = shift;
	
	$$albumHTML =~ /release-date">.*?span>(.*?)</si;
	
	return $1;
}

sub getAlbumRating {
	my $albumHTML = shift;
	my $client = shift;
	
	my $rating;

	$$albumHTML =~ /rating-allmusic-(\d+)/i;
	$rating = $1 + 1;
	$rating = $rating / 2;

	if ($rating) {
		$rating = cstring($client, 'PLUGIN_ALBUMREVIEW_RATING') . " $rating " . cstring($client, 'PLUGIN_ALBUMREVIEW_STARS')
	}

	if ($$albumHTML =~ /album-pick">/i) {
		$rating .= ($rating ? ', ' : '') . cstring($client, 'PLUGIN_ALBUMREVIEW_PICK');
	}

	$log->info("AlbumReview (rating) $rating");

	return $rating;
}

sub initStatus {
	my $client = shift;
	$status{$client}{status} = shift;
	$status{$client}{album} = shift;
	$status{$client}{playerReview} = shift;
	$status{$client}{playerReview} ||= ();
	$status{$client}{line} = shift;
	$status{$client}{htmlReview} = shift;
	$status{$client}{images} = shift;
	$status{$client}{links} = shift;
	$status{$client}{notFound} = shift;
	$status{$client}{trackList} = shift;
	$status{$client}{year} = shift;
	$status{$client}{rating} = shift;
	$status{$client}{creditsUrl} = shift;
	$status{$client}{credits} = shift;	
}

sub setStatus {
	my ($client, $tag, $value) = @_;
	$status{$client}{$tag} = $value;
}

sub getStatus {
	my ($client, $tag) = @_;
	return $status{$client}{$tag};
}

sub getAlbumFromAlbumId {
	my $client = shift;
	my $albumId = shift;
	my %album;

	if (defined($albumId)) {
		my $ds    = 'Slim::Schema';
		my $album = $ds->resultset("Album")->find($albumId);

		if ($album) {
			%album = (
				'artist' => $album->contributor->name,
				'album' => cleanupAlbumName($album->title),			
				'coverart' => $album->artwork
			);
		}
	}

	return \%album;
}

sub getAlbumFromSongURL {
	my $client = shift;
	my %album;

	if (my $url = Slim::Player::Playlist::song($client)) {
		$url = $url->url;
		my $ds    = 'Slim::Schema';
		my $track = $ds->objectForUrl($url);

		my ($artist, $album);
		$artist = $track->artist->name if (defined $track->artist);
		$album = $track->album->title if (defined $track->album);

		# We didn't get an artist - maybe it is some music service?
		unless ($album && $artist) {
			if ($track->remote()) {
				my $handler = Slim::Player::ProtocolHandlers->handlerForURL($url);
				if ( $handler && $handler->can('getMetadataFor') ) {
					my $remoteMeta = $handler->getMetadataFor($client, $url);
					$album ||= $remoteMeta->{album};
					$artist ||= $remoteMeta->{artist};
				}
				$log->debug("AlbumReview (get artist's name): giving a music service a try: '$artist', album: '$album'");
			}
		}

		if ($album && $artist) {
			my $coverart;
			if ($track->coverArt('thumb')) {
				$coverart = $track->id();
			}

			$album = cleanupAlbumName($album);			
	
			$log->warn("AlbumReview (get artist's name): '$artist', album: '$album'");
			if ($artist && $album) {
				%album = ('artist' => $artist, 'album' => $album, 'coverart' => $coverart);
				return \%album;
			}
		}
	}
	return \%album;
}

sub cleanupAlbumName {
	my $album = shift;
	
	$log->debug("Cleaning up album name: '$album'");

	# optionally remove everything between () or []... But don't for PG's eponymous first four albums :-)
	$album =~ s/[\(\[].*?[\)\]]//g if ($prefs->get('plugin_albumreview_remove_brackets') && $album !~ /Peter Gabriel \[[1-4]\]/i);
	
	# remove stuff like "CD02", "1 of 2"
	$album =~ s/\b(disc \d+ of \d+)\b//ig;
	$album =~ s/\d+\/\d+//ig;
	$album =~ s/\b(cd\s*\d+|\d+ of \d+|disc \d+)\b//ig;
	# remove trailing non-word characters
	$album =~ s/[\s\W]{2,}$//;
	$album =~ s/\s*$//;

	$log->debug("Album name cleaned up:  '$album'");

	return $album;
}

sub escapeRegExpChars {
	my $value = shift;
	# escape special characters
	$value =~ s/([\\\(\)\[\]\*\?\.\$])/\\$1/g;
	return $value;
}

sub fixBaseURL {
	my $html = shift;

	$html =~ s/href="(.*?(?:allmusic|allmovie)\.com.*?)"/href="$1" target="_blank"/sig;
	$html =~ s/href="\/(.*?)"/href="$baseURL\/$1" target="_blank"/sig;
	$html =~ s/src="\/(.*?)"/src="$baseURL\/$1"/sig;

	return $html;
}


sub cliQuery {
	my $request = shift;
	
	$log->warn("AlbumReview: cliQuery()");
	
	# check this is the correct query.
	if ($request->isNotQuery([['albumreview'], ['get']]) || !$request->client()) {
		$request->setStatusBadDispatch();
		return;
	}
	
	my $client = $request->client();
	$request->setStatusProcessing();

	my $params = $request->getParamsCopy();

	my $review = getCachedReview($client, $params->{'params'} || $params);
	if ((not $review) && (not defined $status{$client}{status})) {
		no strict 'subs';
		loadReview($client, $params, undef, undef, undef, $request);
	}
	elsif ($review) {
		cliPrepareResponse($request);
	}
}

sub cliPrepareResponse {
	my $request = shift;

	$log->info("cliPrepareResponse()");
	my $client = $request->client();
	
	if ($status{$client}{notFound}) {
		$request->addResult('notfound', '1');

		if ($request->getParam('menu')) {
			$request->addResult('window', {
				textarea => $client->string('PLUGIN_ALBUMREVIEW_NOREVIEW'),
			});	
			
			$request->addResult('offset', 0);
			$request->addResult('count', 0);
		}
		
		$request->setStatusDone();
	}
	
	elsif ($request->isQuery([['albumreview']])) {
		_cliPrepareResponse($client, $request);
	}

	elsif ($request->isQuery([['albumcredits']])) {
		Plugins::AlbumReview::Credits::cliPrepareResponse($client, $request);
	}
}

sub _cliPrepareResponse {
	my ($client, $request) = @_;
	
	if ($request->getParam('menu')) {
		my $review = stripCLIResponse($status{$client}{album}, 'html') if $status{$client}{album};
		$review .= stripCLIResponse($status{$client}{artist}, 'html') if $status{$client}{artist};
		$review .= "($status{$client}{year})\n" if $status{$client}{year};
		$review .= stripCLIResponse($status{$client}{rating}, 'html') if $status{$client}{rating};

		# show list of potentially related albums
		if (@{$status{$client}{playerReview}} == 1 && $status{$client}{playerReview}->[0] eq $client->string('PLUGIN_ALBUMREVIEW_NOREVIEW')) {
			
			my %albumsFound;
			my @sortedAlbums;
			while ($status{$client}{htmlReview} =~ /href="([^"]*?)"[^>]*?>(.+?)<\/li/ig) {
				my $link  = $1;
				my $album = $2;
				$link =~ s/\&amp\;/\&/g;
				$album =~ s/<.*?>//i;

				$albumsFound{$album} = $link;
				push @sortedAlbums, $album;
			}

			my $chunkCount = 0;
			foreach my $item (@sortedAlbums) {
				$review = '';
				
				# don't return more than 200 items
				last if $chunkCount == 200;
				
				$request->addResultLoop('item_loop', $chunkCount, 'text', $item);
				$request->addResultLoop('item_loop', $chunkCount, 'actions', {
					go => {
						cmd => [ 'albumreview', 'get' ],
						params => {
							menu => 1,
							allmusicUrl => $albumsFound{$item},
							album => $item,
						}					
					}
				});		
				
				$chunkCount++;
			}

			$request->addResult('offset', 0);
			$request->addResult('count', $chunkCount) if $chunkCount;
		}

		if ($review) {
	
			$review .= "\n" if $review;
			$review .= stripCLIResponse($status{$client}{htmlReview}, 'html');			

			$request->addResult('window', {
				textarea => $review,
			});
			
			$request->addResult('offset', 0);
			$request->addResult('count', 0);
		}

	}
	
	else {
		my $doStrip = $request->getParam('strip');

		$request->addResult('album', stripCLIResponse($status{$client}{album}, $doStrip));
		$request->addResult('artist', stripCLIResponse($status{$client}{status}, $doStrip));
		$request->addResult('albumyear', $status{$client}{year});
		$request->addResult('albumrating', stripCLIResponse($status{$client}{rating}, $doStrip));
		$request->addResult('albumreview', stripCLIResponse($status{$client}{htmlReview}, $doStrip));

#		my $cnt = 0;
#		foreach my $track (keys %{$status{$client}{trackList}}) {
#			foreach (keys %{$track}) {
#				$request->addResultLoop('@loop', $cnt++, $_, $status{$client}{trackList}{$track}{$_});
#			}
#		}
	}

	$request->setStatusDone();
}

sub stripCLIResponse {
	my ($string, $strip) = @_;
	$string = Slim::Utils::Unicode::utf8decode($string);
	
	if ($strip =~ /html/i)	{
		if ($hasModules) {
			my $reviewFormatter = HTML::FormatText->new(leftmargin => 0, rightmargin => 999999);
			$string = $reviewFormatter->format_string($string);
		}
		else {
			$string =~ s/<(([^ >]|\n)*)>//g;;
		}
	}
	elsif ($strip =~ /links/i)	{
		$string =~ s/<\s*[\/]?\s*a [^>]*?>//sig;
	}

	return $string;
}

sub trackInfoHandler {
	my $return = _objInfoHandler( @_, 'track' );
	return $return;
}

sub albumInfoHandler {
	my $return = _objInfoHandler( @_, 'album' );
	return $return;
}

sub searchHandler {
	my ( $client, $tags ) = @_;
	
	my $return = _objInfoHandler( $client, undef, undef, undef, $tags, 'search' );
}

sub _objInfoHandler {
	my ( $client, undef, $obj, undef, $tags, $objType ) = @_;

	my $album_id = $obj ? $obj->id : undef;

	if ( $objType eq 'track') {
		$album_id = $obj->album ? $obj->album->id : undef;
	}

	return {
		type   => 'redirect',
		name   =>  cstring($client, getDisplayName()),

		player => {
			mode => 'Plugins::AlbumReview::Plugin',
			modeParams => {
				album_id => $album_id,
			},
		},

		web  => {
			url => 'plugins/AlbumReview/index.html?album_id=' . $album_id,
		},
		
		jive => {
			actions => {
				go => {
					cmd => [ 'albumreview', 'get' ],
					params => {
						menu => 1,
						album_id => $album_id,
						album => $tags->{search},
					}					
				}
			}
		}
	};
}

1;
